/***************************************************************************************************
 *	XAniRoC for X11 - V1.02
 *	developed 1993 by Christian Zaenker (zaenker@informatik.tu-muenchen.de)
 *
 *      many thanx to:
 *
 *      David L. Crow, IBM Advanced Workstations and Systems, AIX Systems Graphics Development
 *              for the routines to detect EnterNotify and LeaveNotify events on the root window
 *
 *      Stan Kalinowski, Tektronix, Inc., Network Displays Division
 *              for the signal handler routines and the base for the manpage
 *
 *	- xaniroc will change the root window cursor of an X11 screen into a rotating X
 *	- to use xaniroc your system should be able to create cursors with a size of at least 16x16
 *	- how to create: see README file
 *	- for options type 'xaniroc -h'
 *
 ***************************************************************************************************
 */

#include <stdio.h>				/* get standard I/O functions */
#include <X11/Xlib.h>				/* get the X library definitions */
#include <X11/Xutil.h>				/* get the X11 utility definitions */
#include <sys/signal.h>                         /* get signal definitions */
#include <sys/time.h>                           /* get time definitions */
#include <string.h>                             /* get string functions */
#include "bitmaps"                              /* get the bitmaps for the cursor */

#ifdef  SIGNALRETURNSINT
#define SIGVAL int
#else
#define SIGVAL void
#endif

#define NO      0
#define YES     1


int		frame_freq,
		i,
		speed,
		animating,
		transparent;

char    	display_name[50],
		fg_color_name[50],
		bg_color_name[50];

Display		*display;			/* X server connection */
Pixmap		cursor_source[24],		/* pixmaps (bitmaps) for the cursor */
		cursor_mask[24];
XColor		cursor_fg_color,		/* cursor colors */
		cursor_bg_color;
Cursor		cursor[24];			/* cursors */
XEvent		event;				/* events */
GC		gc_xor,				/* GC's */
                gc_clear;


/***************************************************************************************************
 *	function prototypes
 ***************************************************************************************************
 */

void main(int argc, char *argv[]);
void usage(void);
void frame_wait(void);
int  pointer_in_rootwindow(void);
#ifdef  SVR4
SIGVAL cleanup_and_exit(int sig);
#else
SIGVAL cleanup_and_exit(int sig, int code, struct sigcontext *scp, char *addr);
#endif	/* SVR4 */

/***************************************************************************************************
 *	the main function of XAniroc
 ***************************************************************************************************
 */

void main(int argc, char *argv[])
{

/***************************************************************************************************
 *	set default values
 ***************************************************************************************************
 */
	strcpy(fg_color_name, "Black");
	strcpy(bg_color_name, "White");
	strcpy(display_name, "");
	speed = 25;
	animating = NO;
	transparent = NO;

/***************************************************************************************************
 *	handle the command line arguments
 ***************************************************************************************************
 */
	if(argc > 1)
	{
                for(i = 2; i <= argc; i++)
                {
		        if(strcmp(argv[i - 1], "-h") == 0)      /* usage ? */
		        {
			        usage();
			        exit(0);
                        }

		        if(strcmp(argv[i - 1], "-speed") == 0)	/* speed ? */
		        {
			        if((argv[i] == NULL) || ((speed = atoi(argv[i])) <= 0))
			        {				/* no or too small speed value ? */
				        usage();
				        exit(0);
			        }
			        i++;                            /* jump over value */
        		}

		        if(strcmp(argv[i - 1], "-fc") == 0)	/* set foreground ? */
		        {
			        if(argv[i] == NULL)		/* no colorname ? */
			        {
				        usage();
				        exit(0);
			        }

                                strcpy(fg_color_name, argv[i]);	/* set colorname */
                                i++;                            /* jump over value */
        		}

		        if(strcmp(argv[i - 1], "-bc") == 0)	/* set background (border) ? */
		        {
			        if(argv[i] == NULL)		/* no colorname ? */
			        {
				        usage();
				        exit(0);
			        }

                                strcpy(bg_color_name, argv[i]);	/* set colorname */
                                i++;                            /* jump over value */
        		}

		        if(strcmp(argv[i - 1], "-display") == 0)/* set display ? */
		        {
			        if(argv[i] == NULL)		/* no displayname ? */
			        {
				        usage();
				        exit(0);
			        }

                                strcpy(display_name, argv[i]);	/* set displayname */
                                i++;                            /* jump over value */
        		}

		        if(strcmp(argv[i - 1], "-t") == 0)      /* transparent ? */
		        {
				transparent = YES;
                        }

        	}
	}

/***************************************************************************************************
 *	open the chosen X display
 ***************************************************************************************************
 */
	if (!(display = XOpenDisplay(display_name)))
	{
		fprintf (stderr, "%s: Can't open display %s\n", argv[0],
			XDisplayName(display_name));
		exit (1);
	}

/***************************************************************************************************
 *	set the x input (choose enter/leave and motion notify)
 ***************************************************************************************************
 */
	XSelectInput(display, XDefaultRootWindow(display), LeaveWindowMask | EnterWindowMask);

/***************************************************************************************************
 *	create colors for the cursor
 ***************************************************************************************************
 */
	if(XAllocNamedColor(display, DefaultColormap(display, DefaultScreen(display)),
		fg_color_name, &cursor_fg_color, &cursor_fg_color) == 0)
	{
		fprintf(stderr, "%s: Could not allocate color '%s' for cursor foreground.\n",
			argv[0], fg_color_name);
		exit(0);
	}

	if(XAllocNamedColor(display, DefaultColormap(display, DefaultScreen(display)),
		bg_color_name, &cursor_bg_color, &cursor_bg_color) == 0)
	{
		fprintf(stderr, "%s: Could not allocate color '%s' for cursor border.\n",
			argv[0], bg_color_name);
		exit(0);
	}

/***************************************************************************************************
 *	create the pixmaps for the cursor
 ***************************************************************************************************
 */
	for(i = 0; i < 24; i++)
	{
		cursor_source[i] = XCreateBitmapFromData(display, XDefaultRootWindow(display),
			(char *)(x_source_bits + (i * 32)), 16, 16);
		cursor_mask[i] = XCreateBitmapFromData(display, XDefaultRootWindow(display),
			(char *)(x_mask_bits + (i * 32)), 16, 16);
	}

	if(transparent)
	{
                gc_xor = XCreateGC(display,cursor_source[0] , 0, 0);    /* create a GC to xor */
                XSetFunction(display, gc_xor, GXxor);                   /* source and mask */
                XSetForeground(display, gc_xor, WhitePixel(display, DefaultScreen(display)));
                XSetBackground(display, gc_xor, BlackPixel(display, DefaultScreen(display)));

                gc_clear = XCreateGC(display, cursor_source[0] , 0, 0); /* create a GC to clear */
                XSetFunction(display, gc_clear, GXclear);               /* the source */

		for(i = 0; i < 24; i++)
		{
                        /* xor source and mask */
			XCopyPlane(display, cursor_source[i], cursor_mask[i] , gc_xor,
				0, 0, 16, 16, 0, 0, 1);
                        /* clear the source */
			XFillRectangle(display, cursor_source[i], gc_clear, 0, 0, 16, 16);
		}

		XFreeGC(display, gc_xor);				/* free the GC's */
		XFreeGC(display, gc_clear);
	}

/***************************************************************************************************
 *	create the cursors for the root-window
 ***************************************************************************************************
 */

	if(transparent)
	{
		for(i = 0; i < 24; i++)
		{
			cursor[i] = XCreatePixmapCursor(display, cursor_source[i], cursor_mask[i],
				&cursor_fg_color, &cursor_bg_color, 7, 7);
		}
	}

	else
	{
		for(i = 0; i < 24; i++)
		{
			cursor[i] = XCreatePixmapCursor(display, cursor_source[i], cursor_mask[i],
				&cursor_fg_color, &cursor_bg_color, 7, 7);
		}
	}

/***************************************************************************************************
 *	free the pixmaps used to create the cursors
 ***************************************************************************************************
 */
	for(i = 0; i < 24; i++)
	{
		XFreePixmap(display, cursor_source[i]);
		XFreePixmap(display, cursor_mask[i]);
	}

/***************************************************************************************************
 *	set the signals
 ***************************************************************************************************
 */
        (void) signal(SIGINT,  (SIGVAL(*)())cleanup_and_exit);  /* interrupt */
        (void) signal(SIGQUIT, (SIGVAL(*)())cleanup_and_exit);  /* quit */
        (void) signal(SIGTERM, (SIGVAL(*)())cleanup_and_exit);  /* kill */

	signal(SIGALRM, SIG_IGN);                               /* alarm */

/***************************************************************************************************
 *	make the animation
 ***************************************************************************************************
 */
	if(speed == 1)						/* set frame frequency */
		frame_freq = 999990;
	else
		frame_freq = 1000000/speed;

	if(pointer_in_rootwindow()) animating = YES;		/* set initial value */
	i = 0;							/* initial frame number */

	while(YES)
	{
                if(XPending(display) || !animating)
                {
                        XNextEvent(display, &event);
                        switch (event.type)
                        {
                                case LeaveNotify:
                                        animating = NO;
                                        XUndefineCursor(display, XDefaultRootWindow(display));
                                        break;
                                case EnterNotify:
                                        animating = YES;
                                        break;
                        }
                }

                else
                {
			XDefineCursor(display, XDefaultRootWindow(display), cursor[i]);
			frame_wait();
			i++;
			if(i >= 24) i = 0;
		}
	}
} /*end of main */

/***************************************************************************************************
 *	function frame_wait()
				waits the time until the next frame
 ***************************************************************************************************
 */

void frame_wait(void)
{
struct itimerval timer;
struct timeval   timeout;

	timer.it_interval.tv_sec = 0;
	timer.it_interval.tv_usec = frame_freq;
	timer.it_value.tv_sec = 0;
	timer.it_value.tv_usec = frame_freq;
	setitimer(ITIMER_REAL, &timer, NULL);
	XSync(display, False);
	getitimer(ITIMER_REAL, &timer);
	timeout = timer.it_value;
	select(0, NULL, NULL, NULL, &timeout);
}

/***************************************************************************************************
 *	function usage()
				prints out the usage of xarc
 ***************************************************************************************************
 */

void usage(void)
{
	printf("\nxaniroc - animated x on the root window of an X11 display\n");
	printf("---------------------------------------------------------\n");
	printf("questions and help from zaenker@informatik.tu-muenchen.de\n");
        printf("\nVersion: 1.02\n");
	printf("Usage: xaniroc [options]\n");
	printf("Options:\n");
	printf("        -h                      print this helptext\n");
	printf("        -speed <value>	        set rotation speed to value\n");
	printf("                                (default is 25)\n");
	printf("        -fc <colorname>		set the cursor foreground color\n");
	printf("                                (default is Black)\n");
	printf("        -bc <colorname>		set the cursor border color\n");
	printf("                                (default is White)\n");
	printf("        -t			set the cursor foreground to transparent\n");
	printf("        -display <display>	set the display\n");
	printf("                                (default is $DISPLAY)\n\n");
}

/***************************************************************************************************
 *      function pointer_in_rootwindow()
 *                              checks to see if the pointer is in the root window
 *	returns
 * 		1 if pointer is in root window
 *		0 if pointer is not in root window
 ***************************************************************************************************
 */
int pointer_in_rootwindow(void)
{
	Window		root, child;
	int		rx, ry, wx, wy;
	unsigned int 	kb;
	XQueryPointer(display, XDefaultRootWindow(display), &root, &child,
        	&rx, &ry, &wx, &wy, &kb);
	if (child == 0)			/* check if pointer is in the root window */
        	return YES;		/* or in a child window */
	else
		return NO;
}

/***************************************************************************************************
 *	function cleanup_and_exit()
 *				        signal handler that cleans up the cursor
 ***************************************************************************************************
 */
#ifdef  SVR4
SIGVAL cleanup_and_exit(int sig)
#else
SIGVAL cleanup_and_exit(int sig, int code,struct sigcontext *scp, char *addr)
#endif	/* SVR4 */
{
        XUndefineCursor(display, XDefaultRootWindow(display));
        XCloseDisplay(display);
        exit(0);
}





